#include <stdio.h>
#include <mpi.h>

int main(int argc, char *argv[])
{
	MPI_Status status;
	int num, rank, size, tag, next, prev;

	// inicijalizujemo MPI
	MPI_Init(&argc, &argv);
	// trazimo svoj poredak unutar MPI sveta (prvi je 0)
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	// trazimo velicinu MPI sveta
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	// uzimamo da je oznaka programa 201 (moze i neka druga vrednost)
	tag = 201;
	// moduo sluzi da 'kruzimo' po rednim brojevima poredaka u MPI svetu
	// prethodni u MPI svetu, 
	prev = (rank + size - 1) % size;
	// sledeci u MPI svetu
	next = (rank + 1) % size;
	printf("Ovo je proces %d, u MPI svetu se nalazi izmedju procesa %d i %d\n", rank, prev, next);

	// ako smo mi 'master' koji ima pristup konzoli,
	// pitamo za broj krugova koje hocemo da okrenemo po MPI svetu
	if (rank == 0) {
		printf("Unesite broj krugova po MPI svetu: ");
		scanf("%d", &num);

		printf("Proces %d salje broj %d procesu %d\n", rank, num, next);
		// saljemo celi broj sledecem u MPI svetu
		MPI_Send(&num, 1, MPI_INT, next, tag, MPI_COMM_WORLD); 
	}

	// sve dok je broj koji kruzi po MPI svetu veci od 0, 
	// setamo taj broj po MPI svetu
	do {
		// cekamo da primimo podatak od prethodnog u MPI svetu
		MPI_Recv(&num, 1, MPI_INT, prev, tag, MPI_COMM_WORLD, &status);
		printf("Proces %d je upravo primio broj %d od procesa %d\n", rank, num, prev);
		printf("Status: posiljalac %d, oznaka %d, kod greske %d, duzina poruke u bajtovima %d\n", 
			status.MPI_SOURCE, status.MPI_TAG, status.MPI_ERROR, status.count);

		// ako je ovaj proces 0 (master), umanji broj za 1
		if (rank == 0) {
			--num;
			printf("Process 0 (master) je upravo smanjio broj na %d\n", num);
		}

		printf("Proces %d je upravo poslao broj %d procesu %d\n", rank, num, next);
		// saljemo podatak sledecem u MPI svetu
		MPI_Send(&num, 1, MPI_INT, next, tag, MPI_COMM_WORLD);
	} while (num > 0);

	printf("Proces %d je zavrsio sa slanjem i napusta ciklus\n", rank);

	// proces 0 (master) prima poruku od sebi prethodnog (poslednjeg u MPI svetu)
	if (0 == rank) {
		MPI_Recv(&num, 1, MPI_INT, prev, tag, MPI_COMM_WORLD, &status);
		printf("Proces 0 (master) je upravo primio poslednju poruku\n");
		printf("Status: posiljalac %d, oznaka %d, kod greske %d, duzina poruke u bajtovima %d\n", 
			status.MPI_SOURCE, status.MPI_TAG, status.MPI_ERROR, status.count);
	}

	// zavrsavamo MPI
	MPI_Finalize();

	return 0;
}
